from rdflib import Graph
from urllib.parse import urlparse


class Generator:
    def __init__(self, graph: Graph, namespaces=None):
        pass

    def sh_label_gen(self, uri):
        prefix, namespace, name = self.namespaces.compute_qname(uri)
        return prefix + "_" + name

    def uri_validator(self, x):
        try:
            result = urlparse(x)
            return all([result.scheme, result.netloc])
        except Exception:
            return False

    def gen_graph(self, namespace=None, implicit_class_target=False):
        pass



#!/usr/bin/env python
from .shaclgen import data_graph
from .schema import schema

import click
from sys import stdout, stderr
from rdflib import Graph
from rdflib.util import guess_format
from rdflib.namespace import NamespaceManager
from loguru import logger
import json
import importlib.resources


@click.command()
@click.argument("graphs", nargs=-1, type=str)
@click.option(
    "-o",
    "--ontology",
    is_flag=True,
    help="input file(s) or URL(s) is a schema or ontology",
    default=False,
)
@click.option(
    "--output",
    help="output file or - for standard out",
    default="-",
)
@click.option(
    "-s",
    "--serialization",
    "serial",
    help="result graph serialization, default is turtle. example: -s nt",
    default="turtle",
)
@click.option(
    "-p",
    "--prefixes",
    help="optional declaration of namespace prefixes in a json dictionary.",
    default=None,
)
@click.option(
    "-ns",
    "--namespace",
    type=(str, str),
    help="""
    optional shape namespace declaration.
    example: -ns http://www.example.com exam
    """,
)
@click.option(
    "-i",
    "--implicit",
    "implicit_class_target",
    help="use implicit class targets with RDFS",
    default=False,
)
@click.option(
    "--logfile",
    help="Specify a logfile",
    default=None,
)
@click.option(
    "--loglevel",
    help="Specify the standard output loglevel",
    default="INFO",
)
@click.option(
    "--logfilelevel",
    help="Specify the loglevel for the logfile",
    default="DEBUG",
)
def main(
    graphs,
    ontology,
    output,
    serial,
    prefixes,
    namespace,
    implicit_class_target,
    logfile,
    loglevel,
    logfilelevel,
):
    """
    ---------------------------Shaclgen---------------------------

    Shaclgen takes either a data graph(s) or schema(s) as input and generates
    a basic shape file based on the classes and properties present.

    \b
    usage:
        shaclgen [path or url to graph] [optional arguments]
        $ shaclgen webResource-1-0-0.nt -ns www.example.org exam

    \b
    Multiple graphs:
    To load multiple graphs simply list all the graphs one after the other. The
    RDF serialization does not matter.
    example:
        $ shaclgen webResource-1-0-0.nt collection-1-0-0.ttl

    \b
    Shape files from data graphs:
    By default, the input graph is processed as instance graph.

    Shape files from ontologies:
    If the input is a schema or ontology (-o), shaclgen will generate
    a nested shape file: properties with rdfs:domain defined in the ontology
    will be nested within the appropriate NodeShape. rdfs:range definitions
    for XML and rdfs datatypes are included.

    \b
    Serialization options:
        turtle = turtle
        ntriples = nt
        rdfxml = xml
        n3 = n3

    """

    logger.remove()
    logger.add(stderr, level=loglevel)
    if logfile:
        logger.add(logfile, level=logfilelevel)

    source_graph = Graph()

    for graph in graphs:
        source_graph.parse(graph, format=guess_format(graph))

    path = "prefixes/namespaces.json"
    with importlib.resources.path(__name__, path) as prefixes_path:
        namespaces = NamespaceManager(graph=Graph())

        with open(prefixes_path, "r", encoding="utf-8") as fin:
            for prefix, namespace_iri in json.load(fin).items():
                namespaces.bind(prefix, namespace_iri)

    if prefixes:
        with open(prefixes, "r", encoding="utf-8") as fin:
            for prefix, namespace_iri in json.load(fin).items():
                namespaces.bind(prefix, namespace_iri)

    g = None
    if ontology:
        g = schema(source_graph, namespaces)
    else:
        g = data_graph(source_graph, namespaces)
    shape_graph = g.gen_graph(
        namespace=namespace, implicit_class_target=implicit_class_target
    )

    if output == "-":
        print("## shape file generated by SHACLGEN")
        shape_file = stdout.buffer
    else:
        shape_file = output
    shape_graph.serialize(shape_file, format=serial)


if __name__ == "__main__":
    main()



from loguru import logger
from rdflib import Namespace, Literal, Graph
import rdflib
import collections
from rdflib.namespace import RDF, RDFS, SH
from rdflib.namespace import NamespaceManager
from .generator import Generator

SHGEN = Namespace("http://shaclgen/")


class data_graph(Generator):
    def __init__(self, graph: Graph, namespaces=None):
        self.G = graph

        self.CLASSES = collections.OrderedDict()
        self.PROPS = collections.OrderedDict()
        self.OUT = []

        if namespaces:
            self.namespaces = namespaces
        else:
            self.namespaces = NamespaceManager(graph=Graph())
        self.namespaces.bind("sh", SH)
        self.namespaces.bind("shgen", SHGEN)

    def extract_classes(self):
        types_query = "select distinct ?class_ { ?s rdf:type ?class_ }"
        for row in self.G.query(types_query, initNs={"rdf": RDF}):
            logger.debug(f"Class: {row.class_}")
            self.CLASSES[row.class_] = {"label": self.sh_label_gen(row.class_)}

    def extract_props(self):
        prop_query = "select distinct ?prop { ?s ?prop ?o . filter(?prop != rdf:type)}"
        for property_row in self.G.query(prop_query, initNs={"rdf": RDF, "rdfs": RDFS}):
            prop = property_row.prop
            logger.debug(f"Property: {prop}")
            self.PROPS[prop] = {
                "nodekind": None,
                "cardinality": None,
                "classes": [],
                "exceptions": [],
                "label": self.sh_label_gen(prop),
            }

            self.extract_props_subj_types(prop)

            if len(self.PROPS[prop]["classes"]) == 1:
                self.PROPS[prop]["type"] = "unique"
            else:
                self.PROPS[prop]["type"] = "repeat"

            self.extract_props_obj_types(prop)

    def extract_props_subj_types(self, prop):
        logger.debug(f"Property subject types: {prop}")
        prop_subject_type = "select distinct ?class_ {{ ?sub {prop} ?o ; a ?class_ . }}"

        try:
            class_property_result = self.G.query(
                prop_subject_type.format(prop=prop.n3())
            )
            for class_row in class_property_result:
                class_ = class_row.class_
                self.PROPS[prop]["classes"].append(self.CLASSES[class_]["label"])
        except Exception as e:
            logger.error(e)
            self.PROPS[prop]["exceptions"].append(
                {"exception": e, "query": prop_subject_type.format(prop=prop.n3())}
            )

    def extract_props_obj_types(self, prop):
        logger.debug(f"Property object types: {prop}")
        prop_object_type = """
            select distinct ?literal ?dt ?blank ?iri ?class_ {{
                ?s {prop} ?obj .
                optional {{ ?obj a ?class_ }}
                bind(isLiteral(?obj) as ?literal)
                bind(datatype(?obj) as ?dt)
                bind(isBlank(?obj) as ?blank)
                bind(isIRI(?obj) as ?iri)
            }}"""

        try:
            nodekinds = []
            datatypes = []
            objectclasses = []

            property_object_result = self.G.query(
                prop_object_type.format(prop=prop.n3())
            )
            for object_type_row in property_object_result:
                if object_type_row.literal:
                    nodekinds.append("Literal")
                    datatypes.append(object_type_row.dt)
                elif object_type_row.blank:
                    nodekinds.append("BNode")
                elif object_type_row.iri:
                    nodekinds.append("IRI")

                if object_type_row.blank or object_type_row.iri:
                    if object_type_row.class_:
                        objectclasses.append(object_type_row.class_)

            if len(set(nodekinds)) == 1:
                self.PROPS[prop]["nodekind"] = nodekinds[0]
                if nodekinds[0] == "Literal":
                    if len(set(datatypes)) == 1:
                        self.PROPS[prop]["datatype"] = datatypes[0]
                elif nodekinds[0] in ("IRI", "BNode"):
                    self.PROPS[prop]["objectclasses"] = objectclasses
        except Exception as e:
            logger.error(e)
            self.PROPS[prop]["exceptions"].append(
                {"exception": e, "query": prop_object_type.format(prop=prop.n3())}
            )

    def gen_graph(self, namespace=None, implicit_class_target=False):
        logger.info("Start Extraction of the Data Graph")
        logger.info("Classes …")
        self.extract_classes()
        logger.info("Properties …")
        self.extract_props()
        logger.info("Write resulting SHACL Graph …")
        ng = rdflib.Graph(namespace_manager=self.namespaces)

        if namespace is not None:
            if self.uri_validator(namespace[0]):
                uri = namespace[0]
                if namespace[0][-1] not in ["#", "/", "\\"]:
                    uri = namespace[0] + "#"
                EX = Namespace(uri)
                ng.bind(namespace[1], EX)
            else:
                print("##malformed URI, using http://example.org/ instead...")
                EX = Namespace("http://www.example.org/")
                ng.bind("ex", EX)
        else:
            EX = Namespace("http://www.example.org/")
            ng.bind("ex", EX)

        for c in self.CLASSES.keys():
            label = self.CLASSES[c]["label"]
            ng.add((EX[label], RDF.type, SH.NodeShape))
            ng.add((EX[label], SH.targetClass, c))
            ng.add((EX[label], SH.nodeKind, SH.BlankNodeOrIRI))

        for p in self.PROPS.keys():

            ng.add((EX[self.PROPS[p]["label"]], RDF.type, SH.PropertyShape))
            ng.add((EX[self.PROPS[p]["label"]], SH.path, p))

            for exception in self.PROPS[p]["exceptions"]:
                ng.add(
                    (
                        EX[self.PROPS[p]["label"]],
                        SHGEN.exception,
                        Literal(str(exception["exception"])),
                    )
                )
                ng.add(
                    (
                        EX[self.PROPS[p]["label"]],
                        SHGEN.query,
                        Literal(str(exception["query"])),
                    )
                )

            for class_prop in self.PROPS[p]["classes"]:
                ng.add((EX[class_prop], SH.property, EX[self.PROPS[p]["label"]]))
            if self.PROPS[p]["nodekind"] == "IRI":
                ng.add((EX[self.PROPS[p]["label"]], SH.nodeKind, SH.IRI))
            elif self.PROPS[p]["nodekind"] == "BNode":
                ng.add((EX[self.PROPS[p]["label"]], SH.nodeKind, SH.BlankNode))
            elif self.PROPS[p]["nodekind"] == "Literal":
                ng.add((EX[self.PROPS[p]["label"]], SH.nodeKind, SH.Literal))
            if (
                "objectclasses" in self.PROPS[p]
                and len(self.PROPS[p]["objectclasses"]) > 0
            ):
                if len(set(self.PROPS[p]["objectclasses"])) == 1:
                    ng.add(
                        (
                            EX[self.PROPS[p]["label"]],
                            SH["class"],
                            self.PROPS[p]["objectclasses"][0],
                        )
                    )
                else:
                    classNum = 0
                    listnode = EX[
                        self.PROPS[p]["label"] + "-classlist-" + str(classNum)
                    ]
                    ng.add((EX[self.PROPS[p]["label"]], SH["or"], listnode))
                    for objectclass in sorted(set(self.PROPS[p]["objectclasses"])):
                        nextlistnode = EX[
                            self.PROPS[p]["label"] + "-classlist-" + str(classNum)
                        ]
                        alternativenode = EX[
                            self.PROPS[p]["label"] + "-class-" + str(classNum)
                        ]
                        if classNum > 0:
                            ng.add((listnode, RDF.rest, nextlistnode))
                        listnode = nextlistnode
                        classNum += 1
                        ng.add((listnode, RDF.first, alternativenode))
                        ng.add((alternativenode, SH["class"], objectclass))
                    ng.add((listnode, RDF.rest, RDF.nil))
            if "datatype" in self.PROPS[p]:
                ng.add(
                    (EX[self.PROPS[p]["label"]], SH.datatype, self.PROPS[p]["datatype"])
                )

        return ng



from loguru import logger
from rdflib import Graph, Namespace
from rdflib.namespace import XSD, RDF, RDFS, OWL, SH
from rdflib.namespace import NamespaceManager
from rdflib.term import URIRef, Literal, BNode
import collections
from rdflib.collection import Collection
from .generator import Generator

"""
current assumptions:
    - rdfs:range / unionOf - this is only being used on either all Classes or all Datatypes - is this correct?
    - restrictions on a given property do not require domain or range declaration to be present

"""


class schema(Generator):
    def __init__(self, graph: Graph, namespaces=None):
        self.G = graph

        self.CLASSES = collections.OrderedDict()
        self.PROPS = collections.OrderedDict()
        self.REST = collections.OrderedDict()

        self.datatypes = [
            XSD.string,
            XSD.boolean,
            XSD.time,
            XSD.date,
            XSD.dateTime,
            XSD.integer,
            XSD.decimal,
            XSD.nonNegativeInteger,
            XSD.negativeInteger,
            RDFS.Literal,
            XSD.positiveInteger,
            XSD.nonPositiveInteger,
        ]

        if namespaces:
            self.namespaces = namespaces
        else:
            self.namespaces = NamespaceManager(graph=Graph())
        self.namespaces.bind("sh", SH)

    def extract_props(self):
        properties = []
        self.PROPS = {}

        # gather properties
        property_types = [
            OWL.DatatypeProperty,
            OWL.ObjectProperty,
            OWL.AnnotationProperty,
            OWL.TransitiveProperty,
            OWL.FunctionalProperty,
            RDF.Property,
            OWL.InverseFunctionalProperty,
            OWL.SymmetricProperty,
        ]
        for types in property_types:
            for s, p, o in self.G.triples((None, RDF.type, types)):
                properties.append(s)

        for p in sorted(properties):
            self.PROPS[p] = {}

        # gather property values
        count = 0
        for prop in self.PROPS.keys():
            count = count + 1
            s = URIRef(prop)
            self.PROPS[prop]["domain"] = None
            self.PROPS[prop]["domain_union"] = None
            self.PROPS[prop]["range"] = None
            self.PROPS[prop]["range_union"] = None
            self.PROPS[prop]["range_value"] = None

            self.PROPS[prop]["e_prop"] = []
            self.PROPS[prop]["label"] = self.sh_label_gen(prop)
            self.PROPS[prop]["shape_name"] = None
            self.PROPS[prop]["definition"] = None
            self.PROPS[prop]["type"] = []

            for obje in self.G.objects(subject=prop, predicate=RDF.type):
                self.PROPS[prop]["type"].append(obje)
            for sub, pred, ob in self.G.triples((s, RDFS.domain, None)):
                if type(ob) != BNode:
                    self.PROPS[prop]["domain"] = ob
                else:
                    for sub1, pred1, ob1 in self.G.triples((ob, None, None)):

                        if pred1 == OWL.unionOf:
                            c = Collection(self.G, ob1)
                            self.PROPS[prop]["domain_union"] = c

            for sub, pred, ob in self.G.triples((s, RDFS.range, None)):
                if type(ob) != BNode:
                    self.PROPS[prop]["range"] = ob
                else:
                    for sub1, pred1, ob1 in self.G.triples((ob, None, None)):
                        if pred1 == OWL.oneOf:
                            c = Collection(self.G, ob1)
                            self.PROPS[prop]["range_value"] = c
                        if pred1 == OWL.unionOf:
                            c = Collection(self.G, ob1)
                            self.PROPS[prop]["range_union"] = c

            for equal in self.G.objects(subject=s, predicate=OWL.equivalentProperty):
                self.PROPS[prop]["e_prop"].append(equal)

            for defin in self.G.objects(subject=s, predicate=RDFS.comment):
                self.PROPS[prop]["definition"] = defin

            for name in self.G.objects(subject=s, predicate=RDFS.label):

                self.PROPS[prop]["shape_name"] = name
                if self.PROPS[prop]["shape_name"] is None:
                    self.PROPS[prop]["shape_name"] = self.sh_label_gen(prop)

    def extract_classes(self):
        classes = []
        for s, p, o in self.G.triples((None, RDF.type, OWL.Class)):
            if type(s) != BNode:
                classes.append(s)
            else:
                pass
        for s, p, o in self.G.triples((None, RDF.type, RDFS.Class)):
            if type(s) != BNode:
                classes.append(s)
            else:
                pass

        for c in sorted(classes):
            self.CLASSES[c] = {}
        for c in self.CLASSES.keys():
            self.CLASSES[c]["label"] = self.sh_label_gen(c)
            self.CLASSES[c]["definition"] = None
            s = URIRef(c)
            for name in self.G.objects(subject=s, predicate=RDFS.label):
                self.CLASSES[c]["shape_name"] = name
                if self.CLASSES[c]["shape_name"] is None:
                    self.CLASSES[c]["shape_name"] = self.sh_label_gen(c)
            for defin in self.G.objects(subject=s, predicate=RDFS.comment):
                self.CLASSES[c]["definition"] = defin

    def extract_restrictions(self):
        """
        need equivalent classes
        """

        restrictions = []
        for s, p, o in self.G.triples((None, OWL.onProperty, None)):
            restriction = s
            for s in self.G.subjects(object=restriction, predicate=None):
                if type(s) != BNode:
                    for o in self.G.objects(
                        subject=restriction, predicate=OWL.onProperty
                    ):
                        if type(o) != BNode:
                            restrictions.append(restriction)
        for r in sorted(restrictions):
            self.REST[r] = {}

        for rest in self.REST.keys():
            for o in self.G.objects(subject=rest, predicate=OWL.onProperty):
                self.REST[rest]["onProp"] = o

            for s in self.G.subjects(object=rest, predicate=None):
                self.REST[rest]["onClass"] = s

            rest_type = []
            rest_val = []
            for s, p, o in self.G.triples((rest, OWL.maxCardinality, None)):
                rest_type.append(p)
                rest_val.append(o)
            for p in self.G.triples((rest, OWL.minCardinality, None)):
                rest_type.append(p)
                rest_val.append(o)
            for s, p, o in self.G.triples((rest, OWL.cardinality, None)):
                rest_type.append(p)
                rest_val.append(o)
            for s, p, o in self.G.triples((rest, OWL.allValuesFrom, None)):
                rest_type.append(p)
                rest_val.append(o)
            for s, p, o in self.G.triples((rest, OWL.someValuesFrom, None)):
                rest_type.append(p)
                rest_val.append(o)
            for s, p, o in self.G.triples((rest, OWL.hasValue, None)):
                rest_type.append(p)
                rest_val.append(o)
            for s, p, o in self.G.triples((rest, OWL.minQualifiedCardinality, None)):
                # TODO add onClass or onDataRange
                rest_type.append(p)
                rest_val.append(o)
            for s, p, o in self.G.triples((rest, OWL.maxQualifiedCardinality, None)):
                # TODO add onClass or onDataRange
                rest_type.append(p)
                rest_val.append(o)
            for s, p, o in self.G.triples((rest, OWL.qualifiedCardinality, None)):
                # TODO add onClass or onDataRange
                rest_type.append(p)
                rest_val.append(o)
            self.REST[rest]["type"] = rest_type[0]
            self.REST[rest]["value"] = rest_val[0]

    def gen_graph(self, namespace=None, implicit_class_target=False):
        logger.info("Start Extraction of the Ontology")
        logger.info("Properties …")
        self.extract_props()
        logger.info("Classes …")
        self.extract_classes()
        self.extract_restrictions()
        logger.info("Write resulting SHACL Graph …")
        ng = Graph(namespace_manager=self.namespaces)

        if namespace is not None:
            if self.uri_validator(namespace[0]):
                uri = namespace[0]
                if namespace[0][-1] not in ["#", "/", "\\"]:
                    uri = namespace[0] + "/"
                EX = Namespace(uri)
                ng.bind(namespace[1], EX)
            else:
                print("##malformed URI, using http://example.org/ instead...")
                EX = Namespace("http://www.example.org/")
                ng.bind("ex", EX)
        else:
            EX = Namespace("http://www.example.org/")
            ng.bind("ex", EX)

        # add class Node Shapes
        for c in self.CLASSES.keys():
            subject = c
            clabel = self.CLASSES[c]["label"]

            if not implicit_class_target:
                subject = EX[clabel]
                ng.add((subject, SH.targetClass, c))
            else:
                ng.add((subject, RDF.type, RDFS.Class))
                # Copy rdfs:subClassOf
                for t in self.G.triples((subject, RDFS.subClassOf, None)):
                    ng.add(t)

            ng.add((subject, RDF.type, SH.NodeShape))
            # ng.add((EX[clabel], SH.name, Literal(self.CLASSES[c]['shape_name']+' Node shape')))
            ng.add((subject, SH.nodeKind, SH.BlankNodeOrIRI))
            if self.CLASSES[c]["definition"] is not None:
                ng.add(
                    (subject, SH.description, Literal((self.CLASSES[c]["definition"])))
                )

        for p in self.PROPS.keys():
            label = self.PROPS[p]["label"]
            # ng.add((EX[label], SH.name, Literal(str(self.PROPS[p]['shape_name']) +' Property shape')))
            # copy rdfs:label as property shape names
            for o in self.G.objects(subject=p, predicate=RDFS.label):
                ng.add((EX[label], SH.name, o))
            ng.add((EX[label], RDF.type, SH.PropertyShape))
            ng.add((EX[label], SH.path, p))

            if OWL.FunctionalProperty in self.PROPS[p]["type"]:
                ng.add((EX[label], SH.maxCount, Literal(1)))

            if OWL.InverseFunctionalProperty in self.PROPS[p]["type"]:

                ng.add((EX[label], SH.path, BNode(p + "inverse")))
                ng.add((BNode(p + "inverse"), SH.inversePath, p))
                ng.add((BNode(p + "inverse"), SH.maxCount, Literal(1)))

            if self.PROPS[p]["range_value"] is not None:
                rang = self.PROPS[p]["range_value"]
                st = BNode()
                ng.add((EX[label], SH["in"], st))
                Collection(ng, st, [Literal(x) for x in rang])

            if self.PROPS[p]["range"] is not None:
                rang = self.PROPS[p]["range"]
                if rang in self.datatypes:
                    ng.add((EX[label], SH.datatype, rang))

                else:
                    ng.add((EX[label], SH["class"], rang))

            if self.PROPS[p]["e_prop"] is not None:
                for x in self.PROPS[p]["e_prop"]:
                    ng.add((EX[label], SH.equals, x))

            # create range unions using sh:or
            if self.PROPS[p]["range_union"] is not None:
                rang = self.PROPS[p]["range_union"]
                if set(rang).issubset(self.datatypes):

                    st = BNode(label + str(0) + "a")
                    ng.add((EX[label], EX["or"], st))

                    for x, y in enumerate(rang):
                        if x == 0:
                            ng.add((st, RDF.first, BNode(label + str(x) + "_name")))
                            ng.add((BNode(label + str(x) + "_name"), SH["datatype"], y))

                            ng.add((st, RDF.rest, BNode(label + str(x + 1) + "a")))

                        else:
                            ng.add(
                                (
                                    BNode(label + str(x) + "a"),
                                    RDF.first,
                                    BNode(label + str(x) + "_name"),
                                )
                            )
                            ng.add((BNode(label + str(x) + "_name"), SH["datatype"], y))
                        if x + 1 == len(rang):
                            ng.add((BNode(label + str(x) + "a"), RDF.rest, RDF.nil))
                        else:
                            ng.add(
                                (
                                    BNode(label + str(x) + "a"),
                                    RDF.rest,
                                    BNode(label + str(x + 1) + "a"),
                                )
                            )

                else:

                    st = BNode(label + str(0) + "a")
                    ng.add((EX[label], EX["or"], st))

                    for x, y in enumerate(rang):
                        if x == 0:
                            ng.add((st, RDF.first, BNode(label + str(x) + "_name")))
                            ng.add((BNode(label + str(x) + "_name"), SH["class"], y))

                            ng.add((st, RDF.rest, BNode(label + str(x + 1) + "a")))

                        else:
                            ng.add(
                                (
                                    BNode(label + str(x) + "a"),
                                    RDF.first,
                                    BNode(label + str(x) + "_name"),
                                )
                            )
                            ng.add((BNode(label + str(x) + "_name"), SH["class"], y))
                        if x + 1 == len(rang):
                            ng.add((BNode(label + str(x) + "a"), RDF.rest, RDF.nil))
                        else:
                            ng.add(
                                (
                                    BNode(label + str(x) + "a"),
                                    RDF.rest,
                                    BNode(label + str(x + 1) + "a"),
                                )
                            )

            if self.PROPS[p]["definition"] is not None:
                ng.add(
                    (EX[label], SH.description, Literal((self.PROPS[p]["definition"])))
                )

            if self.PROPS[p]["domain"] is not None:
                subject = self.PROPS[p]["domain"]
                if subject in self.CLASSES.keys():
                    plabel = self.PROPS[p]["label"]
                    if implicit_class_target:
                        ng.add((subject, SH.property, EX[plabel]))
                    else:
                        dlabel = self.CLASSES[subject]["label"]
                        ng.add((EX[dlabel], SH.property, EX[plabel]))

            if self.PROPS[p]["domain_union"] is not None:
                for d in self.PROPS[p]["domain_union"]:
                    if d in self.CLASSES.keys():
                        plabel = self.PROPS[p]["label"]

                        if implicit_class_target:
                            ng.add((d, SH.property, EX[plabel]))
                        else:
                            dlabel = self.CLASSES[d]["label"]
                            ng.add((EX[dlabel], SH.property, EX[plabel]))

        for r in self.REST.keys():
            blank = BNode()

            # if self.REST[r]['onProp'] == p: #and self.REST[r]['onClass'] == self.PROPS[p]['domain']:

            ng.add((EX[self.sh_label_gen(self.REST[r]["onClass"])], SH.property, blank))
            ng.add((blank, SH.path, self.REST[r]["onProp"]))
            if self.REST[r]["type"] in [OWL.cardinality]:

                ng.add(
                    (
                        blank,
                        SH.minCount,
                        Literal(self.REST[r]["value"], datatype=XSD.integer),
                    )
                )
                ng.add(
                    (
                        blank,
                        SH.maxCount,
                        Literal(self.REST[r]["value"], datatype=XSD.integer),
                    )
                )
            elif self.REST[r]["type"] in [OWL.minCardinality]:
                ng.add(
                    (
                        blank,
                        SH.minCount,
                        Literal(self.REST[r]["value"], datatype=XSD.integer),
                    )
                )
            elif self.REST[r]["type"] in [OWL.maxCardinality]:
                ng.add(
                    (
                        blank,
                        SH.maxCount,
                        Literal(self.REST[r]["value"], datatype=XSD.integer),
                    )
                )

            elif self.REST[r]["type"] in [OWL.allValuesFrom]:
                if type(self.REST[r]["value"]) == BNode:

                    for sub1, pred1, ob1 in self.G.triples(
                        (self.REST[r]["value"], None, None)
                    ):
                        if pred1 == OWL.unionOf:
                            union_c = Collection(self.G, ob1)
                            dummy = r + self.REST[r]["value"]
                            nest = BNode(dummy + str(0) + "a")
                            ng.add((blank, SH["or"], nest))
                            for x, y in enumerate(union_c):
                                if x == 0:
                                    ng.add(
                                        (
                                            nest,
                                            RDF.first,
                                            BNode(dummy + str(x) + "_name"),
                                        )
                                    )
                                    ng.add(
                                        (
                                            BNode(dummy + str(x) + "_name"),
                                            SH["class"],
                                            y,
                                        )
                                    )
                                    ng.add(
                                        (
                                            nest,
                                            RDF.rest,
                                            BNode(dummy + str(x + 1) + "a"),
                                        )
                                    )
                                else:
                                    ng.add(
                                        (
                                            BNode(dummy + str(x) + "a"),
                                            RDF.first,
                                            BNode(dummy + str(x) + "_name"),
                                        )
                                    )
                                    ng.add(
                                        (
                                            BNode(dummy + str(x) + "_name"),
                                            SH["class"],
                                            y,
                                        )
                                    )
                                if x == len(rang):
                                    ng.add(
                                        (BNode(dummy + str(x) + "a"), RDF.rest, RDF.nil)
                                    )
                                else:
                                    ng.add(
                                        (
                                            BNode(dummy + str(x) + "a"),
                                            RDF.rest,
                                            BNode(dummy + str(x + 1) + "a"),
                                        )
                                    )
                elif type(self.REST[r]["value"]) in self.datatypes:
                    ng.add((blank, SH["datatype"], self.REST[r]["value"]))
                else:
                    ng.add((blank, SH["class"], self.REST[r]["value"]))

            elif self.REST[r]["type"] in [OWL.someValuesFrom]:
                ng.add(
                    (blank, SH["qualifiedMinCount"], Literal(1, datatype=XSD.integer))
                )
                ng.add((blank, SH["qualifiedValueShape"], BNode("count" + str(r))))
                ng.add((BNode("count" + str(r)), SH["class"], self.REST[r]["value"]))
            else:
                pass

        return ng
